// https://wiki.binefa.cat

#include "PCF8574.h" // https://github.com/RobTillaart/Arduino/tree/master/libraries/PCF8574
#include <PubSubClient.h>
#include <WiFi.h>
#include <Wire.h>
//#include "credentials.h"

#include <Adafruit_ADS1015.h>
#include <Adafruit_BME280.h>

Adafruit_ADS1115 ads;  /* 16-bit version */

#define I2C_SDA 21
#define I2C_SCL 22

#define I2C_SDA_OLED 4
#define I2C_SCL_OLED 15

#define R1 0x01
#define R2 0x02
#define R3 0x04
#define R4 0x08

#define EXTERNAL_BUTTON_B1 27
#define EXTERNAL_BUTTON_B2 19 /* B2 <-> B3) */
#define EXTERNAL_BUTTON_B3 18 /* B2 <-> B3) */
#define EXTERNAL_BUTTON_B4 23

#define R1_3  32
#define R2_3  33
#define R3_3  25
#define R4_3  26

#define EXTERNAL_BUTTON 23
#define ADC_POT 35

PCF8574 PCF_38(0x38);
//PCF8574 PCF_38(0x20);

const char* ssid     = "JESUITESFP_P5";
const char* password = "internetcoses";

const char* mqtt_server = "popotamo.binefa.cat";
const int mqtt_port = 1888; // normally 1883

const char* mqtt_user = "ecat";
const char* mqtt_password = "clotClot";

#define TEMA_PUBLICA_ESTAT_BOTO "/jordi/witty/boto"
#define TEMA_SUBSCRIPCIO_ESTAT_LEDS "/jordi/witty/leds"
#define TEMA_SUBSCRIPCIO_ORDRES "/jordi/witty/ordres"
#define TEMA_PUBLICA_RESPOSTA_LDR "/jordi/witty/resposta/ldr"

WiFiClient espClient;
PubSubClient client(espClient);

static byte byState;
const byte button = EXTERNAL_BUTTON;

void vDelayESP(unsigned long ulMilliseconds) {
  unsigned long ulPreviousMillis = millis();

  do {
    yield();
  } while (millis() - ulPreviousMillis <= ulMilliseconds);
}

void callback(char* topic, byte* payload, unsigned int length) {
  String szRx = "", szTema(topic);
  char sz[33];

  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.print("] ");
  for (int i = 0; i < length; i++) {
    char receivedChar = (char)payload[i];
    szRx += receivedChar;
  }
  Serial.println(szRx);
  if (szTema == TEMA_SUBSCRIPCIO_ORDRES) {
    int16_t adc0, adc1, adc2, adc3;

    adc0 = ads.readADC_SingleEnded(0);
    adc1 = ads.readADC_SingleEnded(1);
    adc2 = ads.readADC_SingleEnded(2);
    adc3 = ads.readADC_SingleEnded(3);
    adc0 = ads.readADC_SingleEnded(0);
    if (szRx == "A2" || szRx == "a2" ) {
      sprintf(sz,"%d",adc2);
      client.publish(TEMA_PUBLICA_RESPOSTA_LDR, sz);
    }
  }
  if (szTema == TEMA_SUBSCRIPCIO_ESTAT_LEDS) {
    if (szRx == "2L" || szRx == "R1L" ) byState &= ~R1;
    if (szRx == "2H" || szRx == "R1H") byState |= R1;
    if (szRx == "12L" || szRx == "R2L") byState &= ~R2;
    if (szRx == "12H" || szRx == "R2H") byState |= R2;
    if (szRx == "13L" || szRx == "R3L") byState &= ~R3;
    if (szRx == "13H" || szRx == "R3H") byState |= R3;;
    if (szRx == "15L" || szRx == "R4L" ) byState &= ~R4;
    if (szRx == "15H" || szRx == "R4H") byState |= R4;
    PCF_38.write8(~byState);
  }
}


void reconnect() {
  // Loop until we're reconnected
  String clientId = "ESP32Client Jordi 20191203";
  while (!client.connected()) {
    Serial.print("Attempting MQTT connection...");
    // Attempt to connect
    if (client.connect(clientId.c_str(), mqtt_user, mqtt_password)) {
      Serial.println("connected");
      // ... and subscribe to topic
      client.subscribe(TEMA_SUBSCRIPCIO_ESTAT_LEDS);
      client.subscribe(TEMA_SUBSCRIPCIO_ORDRES);  
    } else {
      Serial.print("failed, rc=");
      Serial.print(client.state());
      Serial.println(" try again in 5 seconds");
      // Wait 5 seconds before retrying
      vDelayESP(5000);
    }
  }
}

void setup_wifi() {
  WiFi.enableSTA(true);
  vDelayESP(2000);
  WiFi.begin(ssid, password);

  Serial.print("[Connecting]");
  Serial.print(ssid);
  int tries = 0;
  while (WiFi.status() != WL_CONNECTED) {
    vDelayESP(500);
    Serial.print(".");
    //tries++;
    //if (tries > 30) {
    //  break;
    //}
  }
  Serial.println();

  printWifiStatus();

  Serial.println("Connected to wifi");
}

void vPrintAdcValues() {
  int16_t adc0, adc1, adc2, adc3;

  adc0 = ads.readADC_SingleEnded(0);
  adc1 = ads.readADC_SingleEnded(1);
  adc2 = ads.readADC_SingleEnded(2);
  adc3 = ads.readADC_SingleEnded(3);
  adc0 = ads.readADC_SingleEnded(0);
  Serial.print("AIN0: "); Serial.println(adc0);
  Serial.print("AIN1: "); Serial.println(adc1);
  Serial.print("AIN2: "); Serial.println(adc2);
  Serial.print("AIN3: "); Serial.println(adc3);
}

void vSetupADS1115() {
  // ads.setGain(GAIN_TWOTHIRDS);  // 2/3x gain +/- 6.144V  1 bit = 3mV      0.1875mV (default)
  ads.setGain(GAIN_ONE);        // 1x gain   +/- 4.096V  1 bit = 2mV      0.125mV
  // ads.setGain(GAIN_TWO);        // 2x gain   +/- 2.048V  1 bit = 1mV      0.0625mV
  // ads.setGain(GAIN_FOUR);       // 4x gain   +/- 1.024V  1 bit = 0.5mV    0.03125mV
  // ads.setGain(GAIN_EIGHT);      // 8x gain   +/- 0.512V  1 bit = 0.25mV   0.015625mV
  // ads.setGain(GAIN_SIXTEEN);    // 16x gain  +/- 0.256V  1 bit = 0.125mV  0.0078125mV

  ads.begin();
}

void setup() {
  pinMode(EXTERNAL_BUTTON_B1, INPUT_PULLUP);
  pinMode(EXTERNAL_BUTTON_B2, INPUT_PULLUP);
  pinMode(EXTERNAL_BUTTON_B3, INPUT_PULLUP);
  pinMode(EXTERNAL_BUTTON_B4, INPUT_PULLUP);

  pinMode(R1_3, OUTPUT);
  pinMode(R2_3, OUTPUT);
  pinMode(R3_3, OUTPUT);
  pinMode(R4_3, OUTPUT);

  pinMode(EXTERNAL_BUTTON, INPUT_PULLUP);
  Serial.begin(115200);
  Wire.begin(I2C_SDA, I2C_SCL);
  //Wire.begin(I2C_SDA_OLED, I2C_SCL_OLED);

  byState = 0x00;
  PCF_38.write8(~byState);
  vSetupADS1115();
  setup_wifi();
  client.setServer(mqtt_server, mqtt_port);
  client.setCallback(callback);
}

void loop() {
  boolean bButtonState = !digitalRead(EXTERNAL_BUTTON);
  static boolean bLastButtonState = bButtonState;

  static int nCmpt = 0;

  if (nCmpt > 1000) {
    nCmpt = 0;
    digitalWrite(R1_3, !digitalRead(EXTERNAL_BUTTON_B1));
    digitalWrite(R2_3, !digitalRead(EXTERNAL_BUTTON_B2));
    digitalWrite(R3_3, !digitalRead(EXTERNAL_BUTTON_B3));
    digitalWrite(R4_3, !digitalRead(EXTERNAL_BUTTON_B4));
  }
  nCmpt++;

  if (!client.connected()) {
    reconnect();
  }

  if (bButtonState != bLastButtonState) {
    vPrintAdcValues();
    bLastButtonState = bButtonState;
    vDelayESP(50);
    if (!bButtonState)
      client.publish(TEMA_PUBLICA_ESTAT_BOTO, "Botó GPIO4 NO premut");
    else {
      client.publish(TEMA_PUBLICA_ESTAT_BOTO, "Botó GPIO4 ÉS premut");
    }
  }
  client.loop();
}

void printWifiStatus() {
  // print the SSID of the network you're attached to:
  Serial.print("SSID: ");
  Serial.println(WiFi.SSID());

  // print your WiFi ESP8266's IP address:
  IPAddress ip = WiFi.localIP();
  Serial.print("IP Address: ");
  Serial.println(ip);
}
